@@ -23,6 +23,7 @@ from ConfigParser import ConfigParser
 
 import whisper
 from carbon import log
+from carbon.exceptions import CarbonConfigException
 
 from twisted.python import usage
 
@@ -31,23 +32,32 @@ defaults = dict(
   USER="",
   MAX_CACHE_SIZE=float('inf'),
   MAX_UPDATES_PER_SECOND=500,
+  MAX_UPDATES_PER_SECOND_ON_SHUTDOWN=1000,
   MAX_CREATES_PER_MINUTE=float('inf'),
   LINE_RECEIVER_INTERFACE='0.0.0.0',
   LINE_RECEIVER_PORT=2003,
+  LINE_RECEIVER_BACKLOG=1024,
   ENABLE_UDP_LISTENER=False,
   UDP_RECEIVER_INTERFACE='0.0.0.0',
   UDP_RECEIVER_PORT=2003,
   PICKLE_RECEIVER_INTERFACE='0.0.0.0',
   PICKLE_RECEIVER_PORT=2004,
+  PICKLE_RECEIVER_BACKLOG=1024,
   CACHE_QUERY_INTERFACE='0.0.0.0',
   CACHE_QUERY_PORT=7002,
+  CACHE_QUERY_BACKLOG=1024,
   LOG_UPDATES=True,
+  LOG_CACHE_HITS=True,
+  LOG_CACHE_QUEUE_SORTS=True,
   WHISPER_AUTOFLUSH=False,
   WHISPER_SPARSE_CREATE=False,
+  WHISPER_FALLOCATE_CREATE=False,
   WHISPER_LOCK_WRITES=False,
   MAX_DATAPOINTS_PER_MESSAGE=500,
   MAX_AGGREGATION_INTERVALS=5,
+  FORWARD_ALL=False,
   MAX_QUEUE_SIZE=1000,
+  QUEUE_LOW_WATERMARK_PCT = 0.8,
   ENABLE_AMQP=False,
   AMQP_VERBOSE=False,
   BIND_PATTERNS=['#'],
@@ -58,18 +68,23 @@ defaults = dict(
   MANHOLE_PUBLIC_KEY="",
   RELAY_METHOD='rules',
   REPLICATION_FACTOR=1,
+  DIVERSE_REPLICAS=False,
   DESTINATIONS=[],
   USE_FLOW_CONTROL=True,
   USE_INSECURE_UNPICKLER=False,
   USE_WHITELIST=False,
   CARBON_METRIC_PREFIX='carbon',
   CARBON_METRIC_INTERVAL=60,
+  CACHE_WRITE_STRATEGY='sorted',
+  WRITE_BACK_FREQUENCY=None,
+  ENABLE_LOGROTATION=True,
+  LOG_LISTENER_CONNECTIONS=True,
+  AGGREGATION_RULES='aggregation-rules.conf',
+  REWRITE_RULES='rewrite-rules.conf',
+  RELAY_RULES='relay-rules.conf',
 )
 
 
-def _umask(value):
-    return int(value, 8)
-
 def _process_alive(pid):
     if exists("/proc"):
         return exists("/proc/%d" % pid)
@@ -89,21 +104,24 @@ class OrderedConfigParser(ConfigParser):
   _ordered_sections = []
 
   def read(self, path):
-    result = ConfigParser.read(self, path)
+    # Verifies a file exists *and* is readable
+    if not os.access(path, os.R_OK):
+        raise CarbonConfigException("Error: Missing config file or wrong perms on %s" % path)
 
+    result = ConfigParser.read(self, path)
     sections = []
     for line in open(path):
       line = line.strip()
 
       if line.startswith('[') and line.endswith(']'):
-        sections.append( line[1:-1] )
+        sections.append(line[1:-1])
 
     self._ordered_sections = sections
 
     return result
 
   def sections(self):
-    return list( self._ordered_sections ) # return a copy for safety
+    return list(self._ordered_sections)  # return a copy for safety
 
 
 class Settings(dict):
@@ -116,22 +134,22 @@ class Settings(dict):
   def readFrom(self, path, section):
     parser = ConfigParser()
     if not parser.read(path):
-      raise Exception("Failed to read config file %s" % path)
+      raise CarbonConfigException("Failed to read config file %s" % path)
 
     if not parser.has_section(section):
       return
 
-    for key,value in parser.items(section):
+    for key, value in parser.items(section):
       key = key.upper()
 
       # Detect type from defaults dict
       if key in defaults:
-        valueType = type( defaults[key] )
+        valueType = type(defaults[key])
       else:
         valueType = str
 
       if valueType is list:
-        value = [ v.strip() for v in value.split(',') ]
+        value = [v.strip() for v in value.split(',')]
 
       elif valueType is bool:
         value = parser.getboolean(section, key)
@@ -140,10 +158,10 @@ class Settings(dict):
         # Attempt to figure out numeric types automatically
         try:
           value = int(value)
-        except:
+        except ValueError:
           try:
             value = float(value)
-          except:
+          except ValueError:
             pass
 
       self[key] = value
@@ -157,7 +175,7 @@ class CarbonCacheOptions(usage.Options):
 
     optFlags = [
         ["debug", "", "Run in debug mode."],
-        ]
+    ]
 
     optParameters = [
         ["config", "c", None, "Use the given config file."],
@@ -165,7 +183,7 @@ class CarbonCacheOptions(usage.Options):
         ["logdir", "", None, "Write logs to the given directory."],
         ["whitelist", "", None, "List of metric patterns to allow."],
         ["blacklist", "", None, "List of metric patterns to disallow."],
-        ]
+    ]
 
     def postOptions(self):
         global settings
@@ -208,6 +226,12 @@ class CarbonCacheOptions(usage.Options):
             log.msg("Enabling Whisper autoflush")
             whisper.AUTOFLUSH = True
 
+        if settings.WHISPER_FALLOCATE_CREATE:
+            if whisper.CAN_FALLOCATE:
+                log.msg("Enabling Whisper fallocate support")
+            else:
+                log.err("WHISPER_FALLOCATE_CREATE is enabled but linking failed.")
+
         if settings.WHISPER_LOCK_WRITES:
             if whisper.CAN_LOCK:
                 log.msg("Enabling Whisper file locking")
@@ -215,6 +239,12 @@ class CarbonCacheOptions(usage.Options):
             else:
                 log.err("WHISPER_LOCK_WRITES is enabled but import of fcntl module failed.")
 
+        if settings.CACHE_WRITE_STRATEGY not in ('sorted', 'max', 'naive'):
+            log.err("%s is not a valid value for CACHE_WRITE_STRATEGY, defaulting to %s" %
+                    (settings.CACHE_WRITE_STRATEGY, defaults['CACHE_WRITE_STRATEGY']))
+        else:
+            log.msg("Using %s write strategy for cache" %
+                    settings.CACHE_WRITE_STRATEGY)
         if not "action" in self:
             self["action"] = "start"
         self.handleAction()
@@ -229,6 +259,11 @@ class CarbonCacheOptions(usage.Options):
                 logdir = settings.LOG_DIR
                 if not isdir(logdir):
                     os.makedirs(logdir)
+                    if settings.USER:
+                        # We have not yet switched to the specified user,
+                        # but that user must be able to create files in this
+                        # directory.
+                        os.chown(logdir, self.parent["uid"], self.parent["gid"])
                 log.logToDir(logdir)
 
         if self["whitelist"] is None:
@@ -266,7 +301,7 @@ class CarbonCacheOptions(usage.Options):
             try:
                 pid = int(pf.read().strip())
                 pf.close()
-            except:
+            except IOError:
                 print "Could not read pidfile %s" % pidfile
                 raise SystemExit(1)
             print "Sending kill signal to pid %d" % pid
@@ -288,7 +323,7 @@ class CarbonCacheOptions(usage.Options):
             try:
                 pid = int(pf.read().strip())
                 pf.close()
-            except:
+            except IOError:
                 print "Failed to read pid from %s" % pidfile
                 raise SystemExit(1)
 
@@ -306,7 +341,7 @@ class CarbonCacheOptions(usage.Options):
                 try:
                     pid = int(pf.read().strip())
                     pf.close()
-                except:
+                except IOError:
                     print "Could not read pidfile %s" % pidfile
                     raise SystemExit(1)
                 if _process_alive(pid):
@@ -317,7 +352,7 @@ class CarbonCacheOptions(usage.Options):
                     print "Removing stale pidfile %s" % pidfile
                     try:
                         os.unlink(pidfile)
-                    except:
+                    except IOError:
                         print "Could not remove pidfile %s" % pidfile
 
             print "Starting %s (instance %s)" % (program, instance)
@@ -338,12 +373,12 @@ class CarbonAggregatorOptions(CarbonCacheOptions):
     def postOptions(self):
         CarbonCacheOptions.postOptions(self)
         if self["rules"] is None:
-            self["rules"] = join(settings["CONF_DIR"], "aggregation-rules.conf")
+            self["rules"] = join(settings["CONF_DIR"], settings['AGGREGATION_RULES'])
         settings["aggregation-rules"] = self["rules"]
 
         if self["rewrite-rules"] is None:
             self["rewrite-rules"] = join(settings["CONF_DIR"],
-                                         "rewrite-rules.conf")
+                                         settings['REWRITE_RULES'])
         settings["rewrite-rules"] = self["rewrite-rules"]
 
 
@@ -351,17 +386,22 @@ class CarbonRelayOptions(CarbonCacheOptions):
 
     optParameters = [
         ["rules", "", None, "Use the given relay rules file."],
+        ["aggregation-rules", "", None, "Use the given aggregation rules file."],
         ] + CarbonCacheOptions.optParameters
 
     def postOptions(self):
         CarbonCacheOptions.postOptions(self)
         if self["rules"] is None:
-            self["rules"] = join(settings["CONF_DIR"], "relay-rules.conf")
+            self["rules"] = join(settings["CONF_DIR"], settings['RELAY_RULES'])
         settings["relay-rules"] = self["rules"]
 
-        if settings["RELAY_METHOD"] not in ("rules", "consistent-hashing"):
+        if self["aggregation-rules"] is None:
+            self["aggregation-rules"] = join(settings["CONF_DIR"], settings['AGGREGATION_RULES'])
+        settings["aggregation-rules"] = self["aggregation-rules"]
+
+        if settings["RELAY_METHOD"] not in ("rules", "consistent-hashing", "aggregated-consistent-hashing"):
             print ("In carbon.conf, RELAY_METHOD must be either 'rules' or "
-                   "'consistent-hashing'. Invalid value: '%s'" %
+                   "'consistent-hashing' or 'aggregated-consistent-hashing'. Invalid value: '%s'" %
                    settings.RELAY_METHOD)
             sys.exit(1)
 
@@ -372,12 +412,24 @@ def get_default_parser(usage="%prog [options] <start|stop|status>"):
     parser.add_option(
         "--debug", action="store_true",
         help="Run in the foreground, log to stdout")
+    parser.add_option(
+        "--syslog", action="store_true",
+        help="Write logs to syslog")
+    parser.add_option(
+        "--nodaemon", action="store_true",
+        help="Run in the foreground")
     parser.add_option(
         "--profile",
         help="Record performance profile data to the given file")
+    parser.add_option(
+        "--profiler",
+        help="Specify the profiler to use")
     parser.add_option(
         "--pidfile", default=None,
         help="Write pid to the given file")
+    parser.add_option(
+        "--umask", default=None,
+        help="Use the given umask when creating files")
     parser.add_option(
         "--config",
         default=None,
@@ -456,7 +508,7 @@ def read_config(program, options, **kwargs):
     if graphite_root is None:
         graphite_root = os.environ.get('GRAPHITE_ROOT')
     if graphite_root is None:
-        raise ValueError("Either ROOT_DIR or GRAPHITE_ROOT "
+        raise CarbonConfigException("Either ROOT_DIR or GRAPHITE_ROOT "
                          "needs to be provided.")
 
     # Default config directory to root-relative, unless overriden by the
@@ -493,7 +545,7 @@ def read_config(program, options, **kwargs):
     config = options["config"]
 
     if not exists(config):
-        raise ValueError("Error: missing required config %r" % config)
+        raise CarbonConfigException("Error: missing required config %r" % config)
 
     settings.readFrom(config, section)
     settings.setdefault("instance", options["instance"])
@@ -510,7 +562,7 @@ def read_config(program, options, **kwargs):
                  (program, options["instance"])))
         settings["LOG_DIR"] = (options["logdir"] or
                               join(settings["LOG_DIR"],
-                                "%s-%s" % (program ,options["instance"])))
+                                "%s-%s" % (program, options["instance"])))
     else:
         settings["pidfile"] = (
             options["pidfile"] or
